/*
 * $Id: ebdp.c,v 1.3 1996/05/22 19:04:05 root Exp $
 */
#include <sys/types.h>
#include <strings.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#define	FRAMESIZE	2048

#define	FILESIZE	256
char	filename[FILESIZE];

#define	LINESIZE	256
char	linebuf[LINESIZE];
char	linesave[LINESIZE];

FILE	*fp;

#define	MAXARG	8

int	dump(), setfile(), quit();

struct cmdtbl {
	char	*name;
	int	(*func)();
} cmdtbl[] = {
	{ "d",	dump },
	{ "f",	setfile },
	{ "k",	dump },
	{ "q",	quit },
	{ NULL,	NULL },
};

main(argc, argv)
	char	**argv;
{
	int	ac;
	struct	cmdtbl *tp;
	char	*av[MAXARG + 1];

	if (argc > 2) {
		fprintf(stderr, "usage: %s [file]\n", argv[0]);
		exit(1);
	}
	if (argc == 2) {
		if (set_filename(argv[1]) == -1)
			exit(1);
		if ((fp = fopen(filename, "r")) == NULL) {
			perror(filename);
			exit(1);
		}
	}
	printf(": ");
	fflush(stdout);
	while (fgets(linebuf, LINESIZE, stdin) != NULL) {
		if (linebuf[0] == 0 || linebuf[0] == '\n') {
			if (linesave[0] == 0)
				goto next;
			bcopy(linesave, linebuf, LINESIZE);
		}
		else
			bcopy(linebuf, linesave, LINESIZE);
		ac = getargs(linebuf, av, MAXARG);
		if (ac == 0)
			goto next;
		for (tp = cmdtbl; tp->name; tp++) {
			if (strcmp(av[0], tp->name) == 0)
				break;
		}
		if (tp->name == NULL)
			printf("%s: Command not found.\n", av[0]);
		else
			(tp->func)(ac, av);
	next:
		printf(": ");
		fflush(stdout);
	}
}

#define	DUMPSIZE	256

dump(ac, av)
	char	**av;
{
	int	c0, c1, i, j, kanji = 0;
	int	size;
	static	u_long	addr;
	char	*ep;
	u_char	*cp;
	u_char	dumpbuf[DUMPSIZE];

	if (ac > 2) {
		printf("usage: %s [addr]\n", av[0]);
		return;
	}
	if (ac == 2) {
		if (ep = index(av[1], ':')) {
			addr = (strtol(av[1], NULL, 16) - 1) * FRAMESIZE +
			       strtol(ep + 1, NULL, 16);
		}
		else {
			addr = strtol(av[1], &ep, 16);
			if (av[1] == ep) {
				printf("%s: invalid value\n", av[1]);
				return;
			}
			if (*ep == 'p' || *ep == 'P') {
				if (addr == 0) {
					printf("%s: invalid page\n", av[1]);
					return;
				}
				addr = (addr - 1) * FRAMESIZE;
			}
		}
	}
	if (fp == NULL) {
		printf("file isn't set yet.\n");
		return;
	}
	printf("page:offset = %x:%03x\n", addr/FRAMESIZE + 1, addr % FRAMESIZE);
	if (fseek(fp, addr, 0) == -1) {
		perror("fseek");
		return;
	}
	if ((size = fread(dumpbuf, 1, DUMPSIZE, fp)) == 0) {
		if (ferror(fp)) {
			printf("read error\n");
			return;
		}
		printf("EOF\n");
		return;
	}
	if (*av[0] == 'k')
		kanji++;
	cp = dumpbuf;
	for (i = 0; i < size; i++) {
		if (i % 16 == 0)
			printf("%08x  ", addr + i);
		printf("%02x%s", *cp, i % 2 ? " " : "");
		if (i % 16 == 15) {
			putchar(' ');
			for (j = 15; j >= 0; j--) {
				c0 = *(cp - j);
				if (kanji) {
					c1 = *(cp - j + 1);
					if (0x20 < c0 && c0 < 0x75 &&
					    0x20 < c1 && c1 < 0x80) {
						putchar(c0 | 0x80);
						putchar(c1 | 0x80);
					}
					else {
						putchar(isprint(c0) ? c0 : '.');
						putchar(isprint(c1) ? c1 : '.');
					}
					j--;
				}
				else
					putchar(isprint(c0) ? c0 : '.');
			}
			putchar('\n');
		}
		cp++;
	}
	addr += size;
}

quit(ac, av)
	char	**av;
{
	exit(0);
}

setfile(ac, av)
	char	**av;
{
	if (ac != 2) {
		printf("usage: %s file\n", av[0]);
		return;
	}
	if (set_filename(av[1]) == 0) {
		if ((fp = fopen(av[1], "r")) == NULL) {
			perror(av[1]);
			return;
		}
		printf("file is set to %s\n", av[1]);
	}
	return;
}

set_filename(name)
	char	*name;
{
	if (access(name, R_OK) == -1) {
		perror(name);
		return -1;
	}
	if (strlen(name) > FILESIZE - 1) {
		fprintf(stderr, "filename is too long\n");
		return -1;
	}
	strcpy(filename, name);
	return 0;
}

getargs(buf, av, max)
	char	*buf;
	char	**av;
{
	int     c, ac, len;
	char    *cp, *dp;

	len = strlen(buf);
	if (len > 0 && buf[len - 1] == '\n')
		buf[len - 1] = 0;
	cp = buf;
	for (ac = 0; ac < max; ac++) {
		for (; *cp && (*cp == ' ' || *cp == '\t'); cp++)
			;
		if (*cp == 0)
			break;
		*av++ = cp;
		dp = cp;
		while (c = *cp) {
			if (c == '\\') {
				if (*(cp + 1) != 0)
					cp++;
			}
			else if (c == ' ' || c == '\t')
				break;
			*dp++ = *cp++;
		}
		if (*cp == 0)
			*dp = 0;
		else {
			*dp = 0;
			cp++;
		}
	}
	*av = NULL;
	return(ac);
}
