/*
 * Copyright (c) 1995 Noriyuki Shiota,  All rights reserved.
 *
 * $Id: gif.c,v 1.3 1996/05/22 19:04:05 root Exp $
 *
 * Original Author: Akira Satoh <akira@jaist.ac.jp><satoyan@flab.fujitsu.co.jp>
 */
#include <stdio.h>
#include <sys/types.h>
#include "ebs.h"

typedef unsigned int	uint4;

static FILE*   out_fp;         /* output FILE */

#define SETFP(fp)   (out_fp=(fp))
#define PUTS(s)     xputs((s), out_fp)
#define PUT1(c)     xputc((c), out_fp)
#define PUTWORD(w)  (PUT1 ((w) & 0x00FF), PUT1 (((w) >> 8) & 0x00FF))
#define FLUSH()     xflush(out_fp)
#define WRITE(s,n)  xwrite((s), 1, (n), out_fp)

/* Magic Numbers (Hash Table) */
#define HBITS   12
#define HSIZE  5003

static int print_compress_data();
static int output();
static int cl_block();
static int cl_hash();
static int char_init();
static int char_out();
static int flush_char();

int bitmap_gifprint (width, height, off, inp, outp, cmdp)
     int   width;
     int   height;
     int   off;
     FILE  *inp, *outp;
     struct command *cmdp;
{
  xputs("Content-Type: image/giff\n\n", outp);
  height += off;

  SETFP(outp);

  PUTS("GIF89a");

  PUTWORD(width);
  PUTWORD(height);
  
  PUT1(0x88); /* 10001000 */
  PUT1(0x0);
  PUT1(0x0);

  /* -------- global color pallete ------- */

  PUT1(0xFF);			/* Color 0: White */
  PUT1(0xFF);
  PUT1(0xFF);
  PUT1(0x0);			/* Color 1: Brack */
  PUT1(0x0);
  PUT1(0x0);

  /* ----- graphic control extension ----- */
    {
      PUT1(0x21);		/* Introducer */
      PUT1(0xF9);		/* Graphic Control Label */
      PUT1(0x4);
      PUT1(0x01);
      PUTWORD(0);		/* Delay Time */
      PUT1(0x0);		/* Transparent Color Index: Color 0 */
      PUT1(0x0);		/* End */
    }
  /* -------------------------------------- */

  PUT1(0x2c);

  PUTWORD(0);			/* Image Left */
  PUTWORD(0);			/* Image Top */
  PUTWORD(width);		/* Image Width */
  PUTWORD(height);		/* Image Height */
  
  PUT1(0x0);  /* 00000000 */

  /* ------------- image data ------------- */
  PUT1(0x2);			/* LZW Minimum Code Size  */

  print_compress_data (inp, width * height,
			off * (width + 7) / 8, 2+1);

  /* --------- end of image data ---------- */

  PUT1(0x0);

  /*--------------- trailer --------------- */
  PUT1(0x3B);
  unlockfile(cmdp);
  return 0;
}


/*
 *
 * GIF Image compression - modified 'compress'
 *
 * Based on: compress.c - File compression ala IEEE Computer, June 1984.
 *
 * By Authors:  Spencer W. Thomas       (decvax!harpo!utah-cs!utah-gr!thomas)
 *              Jim McKie               (decvax!mcvax!jim)
 *              Steve Davies            (decvax!vax135!petsd!peora!srd)
 *              Ken Turkowski           (decvax!decwrl!turtlevax!ken)
 *              James A. Woods          (decvax!ihnp4!ames!jaw)
 *              Joe Orost               (decvax!vax135!petsd!joe)
 *
 */
static int n_bits;		/* number of bits/code */
static int maxbits = HBITS;	/* user settable max # bits/code */
static int maxcode;		/* maximum code, given n_bits */
static int maxmaxcode
  = (int)1 << HBITS;		/* should NEVER generate this code */

static long htab [HSIZE];
static unsigned short codetab [HSIZE];
#define HASHTABOF(i)       htab [i]
#define CODETABOF(i)       codetab [i]

static int hsize = HSIZE;                 /* for dynamic table sizing */

/*
 * To save much memory, we overlay the table used by compress() with those
 * used by decompress().  The tab_prefix table is the same size and type
 * as the codetab.  The tab_suffix table needs 2**HBITS characters.  We
 * get this from the beginning of htab.  The output stack uses the rest
 * of htab, and contains characters.  There is plenty of room for any
 * possible stack (stack used to be 8000 characters).
 */

static int next_code;	/* first unused entry */

/*
 * block compression parameters -- after all codes are used up,
 * and compression rate changes, start over.
 */
static int clear_flg;

static int offset;
static long int in_count;	/* length of input */
static long int out_count;	/* # of codes output (for debugging) */

/*
 * compress stdin to stdout
 *
 * Algorithm:  use open addressing double hashing (no chaining) on the
 * prefix code / next character combination.  We do a variant of Knuth's
 * algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
 * secondary probe.  Here, the modular division first probe is gives way
 * to a faster exclusive-or manipulation.  Also do block compression with
 * an adaptive reset, whereby the code table is cleared when the compression
 * ratio decreases, but after the table fills.  The variable-length output
 * codes are re-sized at this point, and a special CLEAR code is generated
 * for the decompressor.  Late addition:  construct the table according to
 * file size for noticeable speed improvement on small files.  Please direct
 * questions about this implementation to ames!jaw.
 */

static int g_init_bits;

static int clear_code;
static int end_code;


#define MAXCODE(n)  ((1 << (n)) - 1)

static print_compress_data (inp, size, off, init_bits)
     FILE  *inp;
     long  size, off;
     int   init_bits;
{
  long fcode;
  int i;
  int c;
  int ent;
  int disp;
  int hsize_reg;
  int hshift;

  int bit_p = 0;
  int byte;

#define GETV() ((bit_p ? 0 \
		       : (bit_p = 8, \
			  off > 0 ? (byte = 0, off--) \
				  : (byte = getc(inp)))), \
	        ((byte >> --bit_p) & 0x01))

  /*
   * Set up the globals:  g_init_bits - initial number of bits
   */
  g_init_bits = init_bits;

  /*
   * Set up the necessary values
   */
  offset    = 0;
  out_count = 0;
  clear_flg = 0;
  in_count  = 1;
  n_bits    = g_init_bits;
  maxcode   = MAXCODE(n_bits);
  
  clear_code = (1 << (init_bits - 1));
  end_code   = clear_code + 1;
  next_code = clear_code + 2;

  char_init();

  ent = GETV();

  hshift = 0;
  for (fcode = (long) hsize;  fcode < 65536L; fcode *= 2L)
    ++hshift;

  hshift = 8 - hshift;		/* set hash code range bound */

  hsize_reg = hsize;
  cl_hash ((long)hsize_reg);	/* clear hash table */
  
  output ((int)clear_code);

  while (--size > 0)
    {
      c = GETV();
      ++in_count;

      fcode = (long) (((long) c << maxbits) + ent);
      i = (((int)c << hshift) ^ ent);    /* xor hashing */

      if (HASHTABOF (i) == fcode)
	{
	  ent = CODETABOF (i);
	  continue;
        }
      else if (HASHTABOF (i) < 0 )      /* empty slot */
	goto nomatch;

      disp = hsize_reg - i;           /* secondary hash (after G. Knott) */
      if (i == 0)
	disp = 1;
probe:
      if ((i -= disp) < 0)
	i += hsize_reg;

      if (HASHTABOF (i) == fcode)
	{
	  ent = CODETABOF (i);
	  continue;
        }
      if (HASHTABOF (i) > 0)
	goto probe;

nomatch:
      output ((int) ent);
      ++out_count;
      ent = c;

      if (next_code < maxmaxcode)
	{
	  CODETABOF (i) = next_code++; /* code -> hashtable */
	  HASHTABOF (i) = fcode;
        }
      else
	cl_block();
    }
  /*
   * Put out the final code.
   */
  output ((int)ent);
  ++out_count;
  output ((int)end_code);
}

static uint4 cur_accum;
static int   cur_bits;

static int output (code)
     int  code;
{
  static uint4 masks [] = { 0x0000, 0x0001, 0x0003, 0x0007, 0x000F,
                          0x001F, 0x003F, 0x007F, 0x00FF,
                          0x01FF, 0x03FF, 0x07FF, 0x0FFF,
                          0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF };

  cur_accum &= masks [cur_bits];

    if( cur_bits > 0 )
        cur_accum |= ((long)code << cur_bits);
    else
        cur_accum = code;

    cur_bits += n_bits;

    while (cur_bits >= 8)
     {
        char_out ((unsigned int) (cur_accum & 0x00FF));
        cur_accum >>= 8;
        cur_bits -= 8;
    }

    /*
     * If the next entry is going to be too big for the code size,
     * then increase it, if possible.
     */
   if (next_code > maxcode || clear_flg)
     {

            if( clear_flg )
             {

                maxcode = MAXCODE (n_bits = g_init_bits);
                clear_flg = 0;

             }
            else
             {

                ++n_bits;
                if (n_bits == maxbits)
                    maxcode = maxmaxcode;
                else
                    maxcode = MAXCODE(n_bits);
            }
        }

    if (code == end_code)
      {
        /*
         * At EOF, write the rest of the buffer.
         */
        while (cur_bits > 0)
        {
                char_out ((unsigned int)(cur_accum & 0x00FF));
                cur_accum >>= 8;
                cur_bits -= 8;
        }

        flush_char ();

        FLUSH ();
    }
}

/*
 * Clear out the hash table
 */
static int cl_block ()             /* table clear for block compress */
{
        cl_hash ((long) hsize);
        next_code = clear_code + 2;
        clear_flg = 1;

        output ((int)clear_code);
}

static int cl_hash (hsize)	/* reset code table */
     long hsize;
{

         long *htab_p = htab + hsize;

         long i;
         long m1 = -1;

        i = hsize - 16;
        do {                            /* might use Sys V memset(3) here */
                *(htab_p-16) = m1;
                *(htab_p-15) = m1;
                *(htab_p-14) = m1;
                *(htab_p-13) = m1;
                *(htab_p-12) = m1;
                *(htab_p-11) = m1;
                *(htab_p-10) = m1;
                *(htab_p-9) = m1;
                *(htab_p-8) = m1;
                *(htab_p-7) = m1;
                *(htab_p-6) = m1;
                *(htab_p-5) = m1;
                *(htab_p-4) = m1;
                *(htab_p-3) = m1;
                *(htab_p-2) = m1;
                *(htab_p-1) = m1;
                htab_p -= 16;
        } while ((i -= 16) >= 0);

        for (i += 16; --i >= 0; )
                *--htab_p = m1;
  return 0;
}

/*
 * Number of characters so far in this 'packet'
 */
static int a_count;
/*
 * Define the storage for the packet accumulator
 */
static char accum[ 256 ];

/*
 * Set up the 'byte output' routine
 */
static int char_init()
{
  cur_accum = 0;
  cur_bits = 0;
  
  a_count = 0;
  return 0;
}

/*
 * Add a character to the end of the current packet, and if it is 254
 * characters, flush the packet to disk.
 */
static int char_out (c)
     int c;
{
        accum [a_count++] = c;
        if (a_count >= 254)
                flush_char();
}

/*
 * Flush the packet to disk, and reset the accumulator
 */
static int flush_char ()
{
  if ( a_count > 0 )
    {
    PUT1 (a_count);
    WRITE (accum, a_count);
    a_count = 0;
    }
}
