/*
 * Copyright (c) 1994 Noriyuki Shiota.  All rights reserved.
 *
 * $Id: util.c,v 1.4 1997/05/23 16:08:24 root Exp $
 */
#include <sys/types.h>
#include <sys/file.h>
#include <stdlib.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#include "ebs.h"

usstrcmp(sp1, sp2)
	u_short	*sp1, *sp2;
{
	u_short s1, s2;

	while (s1 = *sp1++) {
		if (s1 != (s2 = *sp2++))
			return s1 - s2;
	}
	return -*sp2;
}

usstrncmp(sp1, sp2, n)
	u_short	*sp1, *sp2;
{
	u_short s1, s2;

	if (n <= 0)
		return 0;
	while (s1 = *sp1++) {
		if (s1 != (s2 = *sp2++) || --n == 0)
			return s1 - s2;
	}
	return -*sp2;
}

usstrlen(str)
	u_short	*str;
{
	u_short *p = str;

	while (*p)
		p++;
	return p - str;
}

u_short *
usstrdup(str)
	u_short	*str;
{
	u_short *sp, *p = str;

	while (*p++)
		;

	p = sp = xmalloc((p - str) * sizeof(u_short));
	while (*p++ = *str++)
		;
	return sp;
}

reverse_key(str)
	u_short	*str;
{
	int	c;
	u_short	*p = str;

	while (*p)
		p++;
	p--;
	while (str < p) {
		c = *str;
		*str++ = *p;
		*p-- = c;
	}
}

struct genqent {
	struct	genqent *next;
};

/*
 *   Generic list reversal
 */
reverse(q)
	struct genqent **q;
{
	int	i;
	struct genqent *p, *p0, *hd = NULL;

	for (i = 0, p = *q; p; p = p0) {
		p0 = p->next;
		p->next = hd;
		hd = p;
		i++;
	}
	*q = hd;
	return i;
}

u_long
get_long(fp)
	FILE	*fp;
{
	u_long	val;

	val = getc(fp) << 24;
	val += getc(fp) << 16;
	val += getc(fp) << 8;
	val += getc(fp);
	return val;
}

u_long
get_short(fp)
	FILE	*fp;
{
	u_long	val;

	val = getc(fp) << 8;
	val += getc(fp);
	return val;
}

u_long
getbcd(fp, len)
	FILE	*fp;
{
	int	i, c, hi, lo;
	u_long	val = 0;

	for (i = 0; i < len; i++) {
		if ((c = getc(fp)) == EOF) {
			fprintf(stderr, "getbcd: unexpected eof\n");
			return 0;
		}
		hi = c >> 4;
		lo = c & 0xf;
		if (hi > 10 || lo > 10) {
			fprintf(stderr, "getbcd: invalid bcd value %x\n", c);
			return 0;
		}
		val *= 100;
		val += hi * 10 + lo;
	}
	return val;
}

u_long
getbinary(fp, len)
	FILE	*fp;
{
	int	c, i;
	u_long	val = 0;

	for (i = 0; i < len; i++) {
		if ((c = getc(fp)) == EOF) {
			fprintf(stderr, "getbcd: unexpected eof\n");
			return 0;
		}
		val = (val << 8) + c;
	}
	return val;
}

get_position(tp, cmdp)
	struct toc *tp;
	struct command *cmdp;
{
	int	len;
	u_long	page, offset;
	u_char	*dp, *cp = cmdp->cpath;

	if (*cp == 0) {
		cmdp->page = tp->tc_start;
		cmdp->offset = 0;
	}
	else {
		if (sscanf(cp, "%x:%x", &page, &offset) != 2)
				return -1;
		cmdp->page = page;
		cmdp->offset = offset;
	}
	cmdp->position = (cmdp->page - 1) * FRAMESIZE + cmdp->offset;
}

char *
current_path(cmdp)
	struct command *cmdp;
{
	int	len;
	static	char path[P_PATHMAX];

	len = cmdp->cpath - cmdp->path;
	bcopy(cmdp->path, path, len);
	path[len] = 0;
	return path;	
}

char *
return_path(cmdp)
	struct command *cmdp;
{
	int	len;
	char	*cp, *endp, *path = cmdp->path;
	static	char pbuf[P_PATHMAX];

	len = strlen(path);
	if (len == 0)
		return "/";
	endp = &path[len - 1];

	cp = endp;
	if (*cp == '/')
		cp--;
	while (path < cp) {
		if (*cp == '/')
			break;
		cp--;
	}
	len = cp - path;
	if (len == 0)
		return "/";

	bcopy(path, pbuf, len + 1);
	pbuf[len + 1] = 0;
	return pbuf;
}

char *
dict_path(cmdp)
	struct command *cmdp;
{
	int	i, len;
	u_char	*cp;
	static	char pbuf[P_PATHMAX];

	/*
	 *   path = "/code/g2/servno/disc/dict/....."
	 */
	i = 6;
	for (cp = cmdp->path; *cp; cp++) {
		if (*cp == '/' && --i == 0)
			break;
	}
	if (i > 0) {
		fprintf(stderr, "dict_path: bad path %s\n", cmdp->path);
		exit(1);
	}
	len = cp - cmdp->path;
	bcopy(cmdp->path, pbuf, len + 1);
	pbuf[len + 1] = 0;
	return pbuf;
}

#ifdef MALLOC_DEBUG
static alloc_count = 0;
#endif

void *
xmalloc(size)
	size_t	size;
{
	void	*p;

#ifdef MALLOC_DEBUG
	alloc_count++;
	fprintf(stderr, "[%d]", alloc_count);
#endif
	if ((p = malloc(size)) == NULL) {
		fprintf(stderr, "memory exausted\n");
		exit(1);
	}
	return p;
}

#ifdef MALLOC_DEBUG
xfree(cp)
	void *cp;
{
	alloc_count--;
	fprintf(stderr, "<%d>", alloc_count);
	free(cp);
}
#endif

xfread(buf, size, nitem, fp, msg)
	char	*buf, *msg;
	FILE	*fp;
{
	int	cc;

	if ((cc = fread(buf, size, nitem, fp)) != nitem) {
		if (feof(fp))
			return 0;
		fprintf(stderr, "%s: ", msg);
		perror("fread");
		return -1;
	}
	return cc;
}

xfseek(fp, off, pos, msg)
	FILE	*fp;
	char	*msg;
{
	if (fseek(fp, off, pos) == -1) {
		fprintf(stderr, "%s: ", msg);
		perror("fseek");
		return -1;
	}
	return 0;
}

/*
 *  줿Ĺ len ʸ fn 
 *  ̾ string Ѵ str 
 */
fname2str(fn, str, len)
	u_char *fn, *str;
{
	int	c, i = 0;
	u_char	*endp = fn + len;

	while (fn < endp) {
		if ((c = *fn++) == ' ' || c == 0)
			break;
		*str++ = tolower(c);
		i++;
	}
	*str = 0;
	return i;
}

fprintkey(fp, key)
	FILE	*fp;
	u_short *key;
{
	u_short c;

	while (c = *key++) {
		putc((c >> 8) | 0x80, fp);
		putc((c & 0xff) | 0x80, fp);
	}
}

lockfile(cmdp, dt)
	struct command *cmdp;
	struct dict *dt;
{
	int f;

#ifdef O_EXLOCK
	if ((f = open(dt->dt_lockpath, O_RDONLY|O_EXLOCK)) == -1) {
		perror("open lockfile");
		exit(1);
	}
#else
	if ((f = open(dt->dt_lockpath, O_RDONLY)) == -1) {
		perror("open lockfile");
		exit(1);
	}
	if (flock(f, LOCK_EX) == -1) {
		perror("flock lockfile");
		exit(1);
	}
#endif
	cmdp->lockf = f;
}

unlockfile(cmdp)
	struct command *cmdp;
{
	close(cmdp->lockf);
}
